 (*****************************************************************************
 *          HOL-TestGen - specification based test case generation
 *                                                                            
 * RBT_test.thy - Testing Red-Black tree Properties
 * Copyright (C) 2004,2005 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-TestGen.                                       
 *                                                                            
 * HOL-TestGen is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-TestGen is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
 * more details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

 
header {* Defining Red-Black Trees *}


theory RBT_def = Testing:




(******************************)

  (* mimicking ml's datatype order *)
text {*
 The implementation of Red-Black
trees is mainly based on the following datatype declaration:
*}
datatype ml_order = LESS | EQUAL | GREATER  

axclass ord_key < type      

consts
  compare :: "'a::ord_key \<Rightarrow> 'a \<Rightarrow> ml_order "  

axclass LINORDER < linorder, ord_key
  LINORDER_less    : "((compare x y) = LESS)    = (x < y)"
  LINORDER_equal   : "((compare x y) = EQUAL)   = (x = y)"
  LINORDER_greater : "((compare x y) = GREATER) = (y < x)"

(* defining red-black trees *)

types    'a item = "'a::ord_key" 

datatype  color = R | B

datatype 'a tree = E | T color "'a tree" "'a item" "'a tree"


  
(* invariants *)

text {*
In this example we have chosen not only to check if keys are stored or
deleted correctly in the trees but also to check if the trees fulfill
the balancing invariants. We formalize the red and black invariant by recursive
predicates:


*}

consts
  isin         :: "'a::LINORDER item \<Rightarrow> 'a tree \<Rightarrow> bool"
  isord        :: "('a::LINORDER item) tree \<Rightarrow> bool"
  weak_redinv  :: "'a tree \<Rightarrow> bool"
  blackinv     :: "'a tree \<Rightarrow> bool"
  redinv:: "'a tree \<Rightarrow> bool"
  max_B_height :: "'a tree \<Rightarrow> nat"

(* test for elements *)
primrec
  isin_empty : "isin x E = False"
  isin_branch: "isin x (T c a y b) = (((compare x y) = EQUAL) 
                                     | (isin x a) | (isin x b))"

(* test for order of elements *) 
primrec
  isord_empty : "isord E = True"
  isord_branch: "isord (T c a y b) 
                = (isord a \<and> isord b 
                \<and> (\<forall> x. isin x a \<longrightarrow> ((compare x y) = LESS))
                \<and> (\<forall> x. isin x b \<longrightarrow> ((compare x y) = GREATER)))"

(* weak red invariant: no red node has a red child *)  
recdef weak_redinv "measure (%t. (size t))"
  weak_redinv_1:   "weak_redinv E = True"
  weak_redinv_2:   "weak_redinv (T B a y b) = (weak_redinv a \<and> weak_redinv b)"
  weak_redinv_3:   "weak_redinv (T R (T R a x b) y c) = False"
  weak_redinv_4:   "weak_redinv (T R a x (T R b y c)) = False"
  weak_redinv_5:   "weak_redinv (T R a x b) = (weak_redinv a \<and> weak_redinv b)"

(* strong red invariant: every red node has an immediate black ancestor, i.e. 
   the root is black and the weak red invariant holds *)
recdef redinv "{}"
  Rinv_1:  "redinv E = True"
  Rinv_2:  "redinv (T R a y b) = False"
  Rinv_3:  "redinv (T B a y b) = (weak_redinv a \<and> weak_redinv b)"

(* calculating maximal number of black nodes on any path from root to leaf *)
recdef max_B_height "measure (%t. (size t))"
  maxB_height_1:  "max_B_height E = 0"
  maxB_height_3:  "max_B_height (T B a y b) 
                   = Suc(max (max_B_height a) (max_B_height b))"
  maxB_height_2:  "max_B_height (T R a y b) 
                   = (max (max_B_height a) (max_B_height b))"

(* black invariant: number of black nodes equal on all paths from root to leaf *)  
recdef blackinv "measure (%t. (size t))"
  blackinv_1:  "blackinv E = True"
  blackinv_2:  "blackinv (T color a y b) 
               = ((blackinv a) \<and> (blackinv b) 
               \<and> ((max_B_height a) = (max_B_height b)))"



instance int::ord_key
  by(intro_classes)

instance int::linorder
  by intro_classes

defs compare_def: "compare (x::int) y 
                  == (if (x < y) then LESS 
                                 else (if (y < x) 
                                       then GREATER 
                                       else EQUAL))"

instance int::LINORDER
apply intro_classes
apply (simp_all add: compare_def)
done


lemma compare1[simp]:"(compare (x::int) y = EQUAL) = (x=y)"
by(auto simp:compare_def)

lemma compare2[simp]:"(compare (x::int) y = LESS) = (x<y)"
by(auto simp:compare_def)

lemma compare3[simp]:"(compare (x::int) y = GREATER) = (y<x)"
by(auto simp:compare_def)


lemma max_0_0 : "((max (a::nat) b) = 0) = (a = 0 \<and> (b = 0))"
  by(auto simp: max_def)

lemma [simp]: "(max (Suc a) b) ~= 0"
  by(auto simp: max_def)

lemma [simp]: "(max b (Suc a) ) ~= 0"
  by(auto simp: max_def)

lemma size_0[simp]: "(size x = 0) = (x = E)"
  by(induct "x",auto)




end
